#ifndef __TEXTFILE_H
#define __TEXTFILE_H

#ifndef __STDIO_H
#include <stdio.h>
#endif                                     

#ifndef __STRINGS_H
#include <strings.h>
#endif

#ifndef __WINDOW_H
#include <window.h>
#endif

#ifndef __RELATION_H
//class TRelation;
#include <relation.h>
#endif

// @doc EXTERNAL

// @enum direction | Direzioni da seguire nella lettura del file di testo
enum direction {
  up,   // @emem Direzione verso l'alto
  down, // @emem Direzione verso il basso
  updown};      // @emem Entrambe le direzioni

// @enum style | Stile del carattere da utilizzare per il disegno del testo
enum style {
  normal = XVT_FS_NONE,                 // @emem Stile carattere normale
  bold = XVT_FS_BOLD,           // @emem Stile carattere in grassetto
  italic = XVT_FS_ITALIC,               // @emem Stile carattere in corsivo
  underlined = XVT_FS_UNDERLINE,        // @emem Stile carattere sottolineato
  tabbed = XVT_FS_SCALE };              // @emem Stile carattere scalato

// @class TTextfile | Classe per la definizione delle procedure di stampa di un testo
//
// @base public | TObject
class TTextfile: public TObject
// @author(:INTERNAL) Villa

// @access Private Member
{
  // @ccost DEFAULT_PAGESIZE | 128 | Dimensione di default della pagina
  enum {DEFAULT_PAGESIZE = 128};

  // @cmember Pagina di testo correntemente in memoria
  TArray _page;
  // @cmember Un bit per ogni riga, TRUE se ci si e' fatto replace
  TBit_array _dirty_lines;
  // @cmember Inizio della pagina in coordinate testo
  long _page_start;
  // @cmember Fine della pagina in coordinate testo
  long _page_end;
  // @cmember Numero di righe di <p _page>
  long _page_size;
  // @cmember Numero totale delle linee di testo
  long _lines;
  // @cmember Numero della linea corrente
  long _cur_line;                             
  // @cmember Nome del file di testo (puo' essere temporaneo)
  TFilename _filename;
  // @cmember Nome del file indice (sempre temporaneo)
  TFilename _indname;
  // @cmember Puntatore al file indice
  FILE* _index;
  // @cmember Puntatore al file di testo
  FILE* _instr;
  // @cmember Direzione ottimale di lettura del file (vedi <t direction>)
  direction _direction;
  
  // @cmember Piece corrente
  int _item;
  // @cmember Testo della riga corrente
  TToken_string _line;
  // @cmember Codifica delllo stile e del colore del carattere
  long _styles[256];
  // @cmember Indica se la piece e' <e style.tabbed>
  bool _tabbed_piece; 

  // @cmember Punti ipertestuali come passati da utente (<c TViswin>)
  TArray _hotspots;
  // @cmember Rappresentazione interna punti ipertestuali
  TArray _spots;
  // @cmember Indice se e' stato modificato
  bool _dirty;
  // @cmember Indica se i file sono aperti
  bool _isopen;
  // @cmember Indica se si tratta di un file temporaneo
  bool _istemp;
  // @cmember Indica se e' permessao dare una <mf TTextfile::append>
  bool _accept;
  // @cmember Indica se si sta modificando con <mf TTextfile::add> e si puo' fare <mf TTextfile::replace>
  bool _interactive;

  // @cmember Gestisce la sostituzione di campi al posto di tag nel testo
  TRelation* _rel;

  // @cmember Riempe la pagina contenente <p line>, ottimizzando start ed end
  void _read_page(long line);
  // @cmember Ritorna TRUE se la roga <p l> e' nella pagina corrente
  bool _in_page(long l)        
  { return  l >= _page_start && l < _page_end; }
  // @cmember Salva i replace fatti sul file (anche temporaneo)
  void _save_changes();
  
  //  void   _parse_style(long j);

  // @cmember Traduce il tag dello stile passato in <p c> (tag) con uno di <t style> (stile XVT)
  style _trans_style(char c);          
  
  // @access Public Member
public:
  // @cmember Ritorna il numero totale delle linee di testo
  long lines()
  { return _lines; }
  // @cmember Ritorna se sono state effettuate delle modifiche
  bool changed() 
  { return _dirty; } 
  

  // @cmember Ritorna la stringa di caratteri senza formattazione
  const char* line(long row, long column = 0, int howmuch = -1);

  // @cmember Ritorna la stringa di caratteri con la formattazione
  const char* line_formatted(long row);

  // @cmember Aggiunge una riga al text (con i formati del caso)
  bool append(const char* l);  

  // @cmember Chiude tutti i files per poter copiare o eseguire operazioni. Dopo aver chiamato
  //          questa funzione non si puo' piu' fare nulla.
  void close();
  // @cmember Stampa su carta
  void print();
  // @cmember Chiude l'aggiunta di nuove linee
  void freeze() 
  { _accept = FALSE; }
  // @cmember Ritorna se e' possibile aggiungere delle nuove righe (TRUE se non vengono accettate)
  bool frozen() 
  { return  !_accept; }
  
  // @cmember Legge il testo formattato
  void read_line(long j, long b = 0, bool pg = TRUE);
  // @cmember Ritorna il pezzo di linea successivo con stile e colore
  const char* piece();            
  // @cmember Ritorna lo stile del piece (vedi <t style>)
  int get_style(int pos = -1);
  // @cmember Ritorna il colore di background del piece
  char get_background(int pos = -1);
  // @cmember Ritorna il colore di foreground del piece
  char get_foreground(int pos = -1); 
  // @cmember Ritorna gli attributi alla posizione <p pos>
  long get_attribute(int pos = -1);

  // @cmember Ritorna la parola alla posizione indicata
  const char* word_at(long x, long y);                  

  // @cmember Ritorna il pezzo di testo indicato (da <p form> a <p to>), allochera' un altro
  //           TText che deve essere disfatto dall'utente.
  TTextfile*  section(TPoint& from, TPoint& to) 
  { return this; }    
  // @cmember Ritorna il nome del file su cui si sta lavorando
  const char* name() 
  { return (const char*)_filename; }  
  // @cmember Permette di settare il valore di <p _iteractive> (vedi sopra)
  void interactive(bool on = TRUE) 
  { _interactive = on; }
  
  // @cmember Scrive il testo (non formattato) su file
  bool write(const char* path, TPoint* from = NULL, TPoint* to = NULL);
  // @cmember Disfa tutto e svuota il file
  void destroy(); 

  // @cmember Cerca una stringa di testo all'interno di una riga
  long search (const char* txt, int& pos, long from = 0, bool down = TRUE, bool casesens = FALSE);
  // @cmember Sostituisce un testo all'interno di una riga
  int replace(long line, const char* txt, int pos = 0, int len = -1);

  // @cmember Permette di settare gli hot_spot
  void set_hotspots(char fg, char bg = 'w');
  // @cmember Ritorna l'array di hotspots
  TArray& hotspots() 
  { return _spots; }
  
  // @cmember Associa una relazione che viene usata (nello stato corrente) per
  //          risolvere eventuali campi per merge (marcati nel testo come
  //          <lt>@file-<gt>fieldname@[format]@[len]@[just]<gt>
  void set_relation(TRelation* r) 
  { _rel = r; }
  
  // @cmember Costruttore
  TTextfile(const char* file = NULL, int pagesize = DEFAULT_PAGESIZE, direction preferred = updown, bool interactive = TRUE);
  // @cmember DIstruttore
  virtual ~TTextfile();
};

#endif

